''' Contains Universal Plugin for phout-compatible shooter '''
from yandextank.plugins.Aggregator import \
    AggregatorPlugin, AggregateResultListener
from yandextank.plugins.Phantom import PhantomReader
from yandextank.core import AbstractPlugin
import subprocess
import time
import shlex
from yandextank.plugins.ConsoleOnline import \
    ConsoleOnlinePlugin, AbstractInfoWidget
import yandextank.plugins.ConsoleScreen as ConsoleScreen
import datetime
from config import PoolConfig, PandoraConfig, parse_schedule
from reader import PandoraReader


class PandoraPlugin(AbstractPlugin, AggregateResultListener):

    '''     Pandora load generator plugin    '''

    OPTION_CONFIG = "config"
    SECTION = "pandora"

    def __init__(self, core):
        super(PandoraPlugin, self).__init__(core)
        self.buffered_seconds = 2
        self.enum_ammo = False
        self.process = None
        self.process_stderr = None
        self.process_start_time = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = ["pandora_cmd", "buffered_seconds",
                "ammo", "loop", "sample_log", "config_file",
                "startup_schedule", "user_schedule",]
        return opts

    def configure(self):
        # plugin part
        self.pandora_cmd = self.get_option("pandora_cmd", "pandora")
        self.buffered_seconds = int(
            self.get_option("buffered_seconds", self.buffered_seconds))

        pool_config = PoolConfig()

        ammo = self.get_option("ammo", "")
        if ammo:
            pool_config.set_ammo(ammo)
        loop_limit = int(self.get_option("loop", "0"))
        pool_config.set_loop(loop_limit)

        self.sample_log = self.get_option("sample_log", "")
        if not self.sample_log:
            self.sample_log = self.core.mkstemp(".samples", "results_")
        self.core.add_artifact_file(self.sample_log)
        pool_config.set_sample_log(self.sample_log)

        startup_schedule = self.get_option("startup_schedule", "")
        if startup_schedule:
            pool_config.set_startup_schedule(parse_schedule(startup_schedule))
        else:
            raise RuntimeError(
                "startup_schedule not specified")

        user_schedule = self.get_option("user_schedule", "")
        if user_schedule:
            pool_config.set_user_schedule(parse_schedule(user_schedule))
        else:
            raise RuntimeError(
                "user_schedule not specified")

        target = self.get_option("target", "localhost:3000")
        pool_config.set_target(target)

        self.pandora_config = PandoraConfig()
        self.pandora_config.add_pool(pool_config)

        self.pandora_config_file = self.get_option("config_file", "")
        if not self.pandora_config_file:
            self.pandora_config_file = self.core.mkstemp(
                ".json", "pandora_config_")
        self.core.add_artifact_file(self.pandora_config_file)
        with open(self.pandora_config_file, 'w') as config_file:
            config_file.write(self.pandora_config.json())

    def prepare_test(self):
        aggregator = None
        try:
            aggregator = self.core.get_plugin_of_type(AggregatorPlugin)
        except KeyError as ex:
            self.log.warning("No aggregator found: %s", ex)

        if aggregator:
            self.log.info(
                "Adding sample reader to aggregator. Reading samples from %s",
                self.sample_log)
            aggregator.reader = PandoraReader(aggregator, self)
            aggregator.reader.buffered_seconds = self.buffered_seconds
            aggregator.add_result_listener(self)
            aggregator.reader.phout_file = self.sample_log

        try:
            console = self.core.get_plugin_of_type(ConsoleOnlinePlugin)
        except KeyError as ex:
            self.log.debug("Console not found: %s", ex)
            console = None

        if console:
            widget = PandoraInfoWidget(self)
            console.add_info_widget(widget)
            aggregator = self.core.get_plugin_of_type(AggregatorPlugin)
            aggregator.add_result_listener(widget)

    def start_test(self):
        args = [self.pandora_cmd, self.pandora_config_file]
        self.log.info("Starting: %s", args)
        self.process_start_time = time.time()
        process_stderr_file = self.core.mkstemp(
            ".log", "pandora_")
        self.core.add_artifact_file(process_stderr_file)
        self.process_stderr = open(process_stderr_file, 'w')
        self.process = subprocess.Popen(
            args, stderr=self.process_stderr,
            stdout=self.process_stderr, close_fds=True)

    def is_test_finished(self):
        retcode = self.process.poll()
        if retcode is not None:
            self.log.info(
                "Subprocess done its work with exit code: %s", retcode)
            return abs(retcode)
        else:
            return -1

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            self.log.warn(
                "Terminating worker process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process_stderr:
                self.process_stderr.close()
        else:
            self.log.debug("Seems subprocess finished OK")
        return retcode

    def get_info(self):
        return None

    def aggregate_second(self, second_aggregate_data):
        pass


class PandoraInfoWidget(AbstractInfoWidget):

    ''' Right panel widget '''

    def __init__(self, pandora):
        AbstractInfoWidget.__init__(self)
        self.krutilka = ConsoleScreen.krutilka()
        self.owner = pandora
        self.rps = 0

    def get_index(self):
        return 0

    def aggregate_second(self, second_aggregate_data):
        self.active_threads = second_aggregate_data.overall.active_threads
        self.rps = second_aggregate_data.overall.RPS

    def render(self, screen):
        text = " Pandora Test %s" % self.krutilka.next()
        space = screen.right_panel_width - len(text) - 1
        left_spaces = space / 2
        right_spaces = space / 2

        dur_seconds = int(time.time()) - int(self.owner.process_start_time)
        duration = str(datetime.timedelta(seconds=dur_seconds))

        template = screen.markup.BG_BROWN + '~' * left_spaces + \
            text + ' ' + '~' * right_spaces + screen.markup.RESET + "\n"
        template += "Command Line: %s\n"
        template += "    Duration: %s\n"
        template += " Responses/s: %s"
        data = (self.owner.pandora_cmd, duration, self.rps)

        return template % data



from yandextank.plugins.Aggregator import AbstractReader
from yandextank.stepper import info as si


class BFGReader(AbstractReader):

    '''
    Listens results from BFG and provides them to Aggregator
    '''

    def __init__(self, aggregator, bfg, result_cache_size=5):
        AbstractReader.__init__(self, aggregator)
        self.bfg = bfg
        self.result_cache_size = result_cache_size
        self.steps = map(list, si.status.get_info().steps)

    def get_next_sample(self, force):
        new_data = []
        while not self.bfg.results.empty():
            new_data.append(self.bfg.results.get(1))
        for cur_time, sample in new_data:
            if not cur_time in self.data_buffer.keys():
                self.data_queue.append(cur_time)
                self.data_buffer[cur_time] = []
            self.data_buffer[cur_time].append(list(sample))
        if self.data_queue and len(self.data_queue) >= self.result_cache_size:
            res = self.pop_second()
            res.overall.planned_requests = self.__get_expected_rps()
            return res
        else:
            return None

    def __get_expected_rps(self):
        '''
        Mark second with expected rps
        '''
        while self.steps and self.steps[0][1] < 1:
            self.steps.pop(0)

        if not self.steps:
            return 0
        else:
            self.steps[0][1] -= 1
            return self.steps[0][0]



'''
Module contains top-level generators.
'''
from itertools import izip
import format as fmt
from config import ComponentFactory
import info
import os
import json
import hashlib
import logging
import re
from util import get_opener


class AmmoFactory(object):

    '''
    A generator that produces ammo.
    '''

    def __init__(self, factory):
        '''
        Factory parameter is a configured ComponentFactory that
        is able to produce load plan and ammo generator.
        '''
        self.factory = factory
        self.load_plan = factory.get_load_plan()
        self.ammo_generator = factory.get_ammo_generator()
        self.filter = factory.get_filter()
        self.marker = factory.get_marker()

    def __iter__(self):
        '''
        Returns a generator of (timestamp, marker, missile) tuples
        where missile is in a string representation. Load Plan (timestamps
        generator) and ammo generator are taken from the previously
        configured ComponentFactory, passed as a parameter to the
        __init__ method of this class.
        '''
        ammo_stream = (ammo for ammo in (
            (missile, marker or self.marker(missile))
            for missile, marker
            in self.ammo_generator
        ) if self.filter(ammo))

        return (
            (timestamp, marker or self.marker(missile), missile)
            for timestamp, (missile, marker)
            in izip(self.load_plan, ammo_stream)
        )

class Stepper(object):

    def __init__(self, **kwargs):
        info.status = info.StepperStatus()
        self.af = AmmoFactory(ComponentFactory(**kwargs))
        self.ammo = fmt.Stpd(self.af)

    def write(self, f):
        for missile in self.ammo:
            f.write(missile)
            try:
                info.status.inc_ammo_count()
            except StopIteration:
                break


class StepperWrapper(object):
    # TODO: review and rewrite this class

    '''
    Wrapper for cached stepper functionality
    '''
    OPTION_STPD = 'stpd_file'
    OPTION_STEPS = 'steps'
    OPTION_TEST_DURATION = 'test_duration'
    OPTION_AMMO_COUNT = 'ammo_count'
    OPTION_LOOP = 'loop'
    OPTION_LOOP_COUNT = 'loop_count'
    OPTION_AMMOFILE = "ammofile"
    OPTION_SCHEDULE = 'rps_schedule'
    OPTION_LOADSCHEME = 'loadscheme'
    OPTION_INSTANCES_LIMIT = 'instances'

    def __init__(self, core, section):
        self.log = logging.getLogger(__name__)
        self.core = core
        self.section = section

        self.cache_dir = '.'

        # per-shoot params
        self.instances = 1000
        self.rps_schedule = []
        self.http_ver = '1.0'
        self.ammo_file = None
        self.instances_schedule = ''
        self.loop_limit = -1
        self.ammo_limit = -1
        self.uris = []
        self.headers = []
        self.autocases = 0
        self.enum_ammo = False
        self.use_caching = True
        self.force_stepping = None
        self.ammo_type = 'phantom'
        self.chosen_cases = []

        # out params
        self.stpd = None
        self.steps = []
        self.ammo_count = 1
        self.duration = 0
        self.loop_count = 0
        self.loadscheme = ""
        self.file_cache = 8192

    def get_option(self, option_ammofile, param2=None):
        ''' get_option wrapper'''
        result = self.core.get_option(self.section, option_ammofile, param2)
        self.log.debug(
            "Option %s.%s = %s", self.section, option_ammofile, result)
        return result

    @staticmethod
    def get_available_options():
        opts = [StepperWrapper.OPTION_AMMOFILE, StepperWrapper.OPTION_LOOP,
                StepperWrapper.OPTION_SCHEDULE, StepperWrapper.OPTION_STPD,
                StepperWrapper.OPTION_INSTANCES_LIMIT]
        opts += ["instances_schedule", "uris",
                 "headers", "header_http",
                 "autocases", "enum_ammo",
                 "ammo_type", "ammo_limit"]
        opts += ["use_caching", "cache_dir",
                 "force_stepping", "file_cache",
                 "chosen_cases"]
        return opts

    def read_config(self):
        ''' stepper part of reading options '''
        self.log.info("Configuring StepperWrapper...")
        self.ammo_file = self.get_option(self.OPTION_AMMOFILE, '')
        self.ammo_type = self.get_option('ammo_type', self.ammo_type)
        if self.ammo_file:
            self.ammo_file = os.path.expanduser(self.ammo_file)
        self.loop_limit = int(self.get_option(self.OPTION_LOOP, "-1"))
        self.ammo_limit = int(self.get_option("ammo_limit", "-1"))

        def make_steps(schedule):
            steps = []
            for step in " ".join(schedule.split("\n")).split(')'):
                if step.strip():
                    steps.append(step.strip() + ')')
            return steps
        self.rps_schedule = make_steps(
            self.get_option(self.OPTION_SCHEDULE, ''))
        self.instances_schedule = make_steps(
            self.get_option("instances_schedule", ''))
        self.instances = int(
            self.get_option(self.OPTION_INSTANCES_LIMIT, '1000'))
        self.uris = self.get_option("uris", '').strip().split("\n")
        while '' in self.uris:
            self.uris.remove('')
        rx = re.compile('\[(.*?)\]')
        self.headers = rx.findall(self.get_option("headers", ''))
        self.http_ver = self.get_option("header_http", self.http_ver)
        self.autocases = self.get_option("autocases", '0')
        self.enum_ammo = self.get_option("enum_ammo", False)
        self.use_caching = int(self.get_option("use_caching", '1'))

        self.file_cache = int(self.get_option('file_cache', '8192'))
        cache_dir = self.core.get_option(
            self.section, "cache_dir", self.core.artifacts_base_dir)
        self.cache_dir = os.path.expanduser(cache_dir)
        self.force_stepping = int(self.get_option("force_stepping", '0'))
        self.stpd = self.get_option(self.OPTION_STPD, "")
        self.chosen_cases = self.get_option("chosen_cases", "").split()
        if self.chosen_cases:
            self.log.info("chosen_cases LIMITS: %s", self.chosen_cases)

    def prepare_stepper(self):
        ''' Generate test data if necessary '''
        def publish_info(stepper_info):
            info.status.publish('loadscheme', stepper_info.loadscheme)
            info.status.publish('loop_count', stepper_info.loop_count)
            info.status.publish('steps', stepper_info.steps)
            info.status.publish('duration', stepper_info.duration)
            info.status.ammo_count = stepper_info.ammo_count
            info.status.publish('instances', stepper_info.instances)
            return stepper_info
        if not self.stpd:
            self.stpd = self.__get_stpd_filename()
            self.core.set_option(self.section, self.OPTION_STPD, self.stpd)
            if self.use_caching and not self.force_stepping and os.path.exists(self.stpd) and os.path.exists(self.__si_filename()):
                self.log.info("Using cached stpd-file: %s", self.stpd)
                stepper_info = self.__read_cached_options()
                if self.instances and self.rps_schedule:
                    self.log.info(
                        "rps_schedule is set. Overriding cached instances param from config: %s", self.instances)
                    stepper_info = stepper_info._replace(instances=self.instances)
                publish_info(stepper_info)
            else:
                if self.force_stepping and os.path.exists(self.__si_filename()):
                    os.remove(self.__si_filename())
                self.__make_stpd_file()
                stepper_info = info.status.get_info()
                self.__write_cached_options(stepper_info)
        else:
            self.log.info("Using specified stpd-file: %s", self.stpd)
            stepper_info = publish_info(self.__read_cached_options())
        self.ammo_count = stepper_info.ammo_count
        self.duration = stepper_info.duration
        self.loop_count = stepper_info.loop_count
        self.loadscheme = stepper_info.loadscheme
        self.steps = stepper_info.steps
        if stepper_info.instances:
            self.instances = stepper_info.instances

    def __si_filename(self):
        '''Return name for stepper_info json file'''
        return "%s_si.json" % self.stpd

    def __get_stpd_filename(self):
        ''' Choose the name for stepped data file '''
        if self.use_caching:
            sep = "|"
            hasher = hashlib.md5()
            hashed_str = "cache version 5" + sep + \
                ';'.join(self.instances_schedule) + sep + str(self.loop_limit)
            hashed_str += sep + str(self.ammo_limit) + sep + ';'.join(
                self.rps_schedule) + sep + str(self.autocases)
            hashed_str += sep + \
                ";".join(self.uris) + sep + ";".join(
                    self.headers) + sep + self.http_ver + sep + ";".join(self.chosen_cases)
            hashed_str += sep + str(self.enum_ammo)
            if self.instances_schedule:
                hashed_str += sep + str(self.instances)
            if self.ammo_file:
                opener = get_opener(self.ammo_file)
                hashed_str += sep + opener.hash
            else:
                if not self.uris:
                    raise RuntimeError(
                        "Neither ammofile nor uris specified")
                hashed_str += sep + \
                    ';'.join(self.uris) + sep + ';'.join(self.headers)
            self.log.debug("stpd-hash source: %s", hashed_str)
            hasher.update(hashed_str)
            if not os.path.exists(self.cache_dir):
                os.makedirs(self.cache_dir)
            stpd = self.cache_dir + '/' + \
                os.path.basename(self.ammo_file) + \
                "_" + hasher.hexdigest() + ".stpd"
        else:
            stpd = os.path.realpath("ammo.stpd")
        self.log.debug("Generated cache file name: %s", stpd)
        return stpd

    def __read_cached_options(self):
        '''
        Read stepper info from json
        '''
        self.log.debug("Reading cached stepper info: %s", self.__si_filename())
        with open(self.__si_filename(), 'r') as si_file:
            si = info.StepperInfo(**json.load(si_file))
        return si

    def __write_cached_options(self, si):
        '''
        Write stepper info to json
        '''
        self.log.debug("Saving stepper info: %s", self.__si_filename())
        with open(self.__si_filename(), 'w') as si_file:
            json.dump(si._asdict(), si_file, indent=4)

    def __make_stpd_file(self):
        ''' stpd generation using Stepper class '''
        self.log.info("Making stpd-file: %s", self.stpd)
        stepper = Stepper(
            rps_schedule=self.rps_schedule,
            http_ver=self.http_ver,
            ammo_file=self.ammo_file,
            instances_schedule=self.instances_schedule,
            instances=self.instances,
            loop_limit=self.loop_limit,
            ammo_limit=self.ammo_limit,
            uris=self.uris,
            headers=[header.strip('[]') for header in self.headers],
            autocases=self.autocases,
            enum_ammo=self.enum_ammo,
            ammo_type=self.ammo_type,
            chosen_cases=self.chosen_cases,
        )
        with open(self.stpd, 'w', self.file_cache) as os:
            stepper.write(os)



"""Target monitoring via SSH"""

import ConfigParser
from collections import defaultdict
from xml.etree import ElementTree as etree
from subprocess import PIPE, Popen
import base64
import logging
import os.path
import re
import select
import signal
import sys
import tempfile
import time
import fcntl
import traceback
import getpass

import yandextank.core as tankcore


logger = logging.getLogger(__name__)


def parse_xml(config):
    if os.path.exists(config):
        return etree.parse(config)
    else:
        return etree.fromstring(config)


class Config(object):
    """Config reader helper"""

    def __init__(self, config):
        self.tree = parse_xml(config)

    def loglevel(self):
        """Get log level from config file. Possible values: info, debug"""
        log_level = 'info'
        log_level_raw = self.tree.getroot().get('loglevel')
        if log_level_raw in ('info', 'debug'):
            log_level = log_level_raw
        return log_level


class SSHWrapper:
    """Separate SSH calls to be able to unit test the collector"""

    def __init__(self, timeout):
        self.ssh_opts = ['-q', '-o', 'StrictHostKeyChecking=no', '-o', 'PasswordAuthentication=no', '-o',
                         'NumberOfPasswordPrompts=0', '-o', 'ConnectTimeout=' + str(timeout)]
        self.scp_opts = []
        self.host = None
        self.port = None

    def set_host_port(self, host, port, username):
        """Set host and port to use"""
        self.host = host
        self.port = port
        self.username = username
        self.scp_opts = self.ssh_opts + ['-P', self.port]
        self.ssh_opts = self.ssh_opts + ['-C', '-p', self.port] + ['-l', self.username]

    def get_ssh_pipe(self, cmd):
        """Get open ssh pipe"""
        args = ['ssh'] + self.ssh_opts + [self.host] + cmd
        logger.debug('Executing: %s', args)
        return Popen(args, stdout=PIPE, stderr=PIPE, stdin=PIPE, bufsize=0, preexec_fn=os.setsid, close_fds=True)

    def get_scp_pipe(self, cmd):
        """Get open scp pipe"""
        args = ['scp'] + self.scp_opts + cmd
        logger.debug('Executing: %s', args)
        return Popen(args, stdout=PIPE, stderr=PIPE, stdin=PIPE, bufsize=0, preexec_fn=os.setsid, close_fds=True)


class AgentClient(object):
    """Agent client connection"""

    def __init__(self):
        self.run = []
        self.host = None

        self.port = 22
        self.ssh = None

        temp_config = tempfile.mkstemp('.cfg', 'agent_')
        os.close(temp_config[0])
        self.path = {
            # Destination path on remote host
            'AGENT_REMOTE_FOLDER': '/var/tmp/lunapark_monitoring',

            # Source path on tank
            'AGENT_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent/',
            'METRIC_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent/metric',

            # Temp config path
            'TEMP_CONFIG': temp_config[1]
        }
        self.interval = None
        self.metric = None
        self.custom = {}
        self.startups = []
        self.shutdowns = []
        self.python = '/usr/bin/env python2'
        self.username = getpass.getuser()

    def start(self):
        """Start remote agent"""
        logger.debug('Start monitoring: %s', self.host)
        if not self.run:
            raise ValueError("Empty run string")
        self.run += ['-t', str(int(time.time()))]
        logger.debug(self.run)
        pipe = self.ssh.get_ssh_pipe(self.run)
        logger.debug("Started: %s", pipe)
        return pipe

    def create_agent_config(self, loglevel):
        """Creating config"""
        try:
            float(self.interval)
        except:
            strn = "Monitoring interval parameter is in wrong format: '%s'. Only numbers allowed."
            raise ValueError(strn % self.interval)

        cfg = ConfigParser.ConfigParser()
        cfg.add_section('main')
        cfg.set('main', 'interval', self.interval)
        cfg.set('main', 'host', self.host)
        cfg.set('main', 'loglevel', loglevel)
        cfg.set('main', 'username', self.username)

        cfg.add_section('metric')
        cfg.set('metric', 'names', self.metric)

        cfg.add_section('custom')
        for method in self.custom:
            if self.custom[method]:
                cfg.set('custom', method, ','.join(self.custom[method]))

        cfg.add_section('startup')
        for idx, cmd in enumerate(self.startups):
            cfg.set('startup', "cmd%s" % idx, cmd)

        cfg.add_section('shutdown')
        for idx, cmd in enumerate(self.shutdowns):
            cfg.set('shutdown', "cmd%s" % idx, cmd)

        with open(self.path['TEMP_CONFIG'], 'w') as fds:
            cfg.write(fds)

        return self.path['TEMP_CONFIG']

    def install(self, loglevel):
        """Create folder and copy agent and metrics scripts to remote host"""
        logger.info("Installing monitoring agent at %s@%s...", self.username, self.host)
        agent_config = self.create_agent_config(loglevel)

        self.ssh.set_host_port(self.host, self.port, self.username)

        # getting remote temp dir
        cmd = [self.python + ' -c "import tempfile; print tempfile.mkdtemp();"']
        logger.debug("Get remote temp dir: %s", cmd)
        pipe = self.ssh.get_ssh_pipe(cmd)

        err = pipe.stderr.read().strip()
        if err:
            raise RuntimeError("[%s] ssh error: '%s'" % (self.host, err))
        pipe.wait()
        logger.debug("Return code [%s]: %s", self.host, pipe.returncode)
        if pipe.returncode:
            raise RuntimeError("Failed to get remote dir via SSH at %s@%s, code %s: %s" % (
                self.username, self.host, pipe.returncode, pipe.stdout.read().strip()))

        remote_dir = pipe.stdout.read().strip()
        if remote_dir:
            self.path['AGENT_REMOTE_FOLDER'] = remote_dir
        logger.debug("Remote dir at %s:%s", self.host, self.path['AGENT_REMOTE_FOLDER'])

        # Copy agent
        cmd = [self.path['AGENT_LOCAL_FOLDER'] + '/agent.py',
               self.username+'@'+'['+self.host+']' + ':' + self.path['AGENT_REMOTE_FOLDER']]
        logger.debug("Copy agent to %s: %s", self.host, cmd)

        pipe = self.ssh.get_scp_pipe(cmd)
        pipe.wait()
        logger.debug("AgentClient copy exitcode: %s", pipe.returncode)
        if pipe.returncode != 0:
            raise RuntimeError("AgentClient copy exitcode: %s" % pipe.returncode)

        # Copy config
        cmd = [
            self.path['TEMP_CONFIG'],
            '{user}@[{host}]:{dirname}/agent.cfg'.format(
                user=self.username,
                host=self.host,
                dirname=self.path['AGENT_REMOTE_FOLDER']
            )
        ]
        logger.debug("[%s] Copy config: %s", cmd, self.host)

        pipe = self.ssh.get_scp_pipe(cmd)
        pipe.wait()
        logger.debug("AgentClient copy config exitcode: %s", pipe.returncode)
        if pipe.returncode != 0:
            raise RuntimeError("AgentClient copy config exitcode: %s" % pipe.returncode)

        if os.getenv("DEBUG") or 1:
            debug = "DEBUG=1"
        else:
            debug = ""
        self.run = [debug, self.python, self.path['AGENT_REMOTE_FOLDER'] + '/agent.py', '-c',
                    self.path['AGENT_REMOTE_FOLDER'] + '/agent.cfg']
        return agent_config

    def uninstall(self):
        """Remove agent's files from remote host"""
        fhandle, log_file = tempfile.mkstemp('.log', "agent_" + self.host + "_")
        os.close(fhandle)
        cmd = [self.host + ':' + self.path['AGENT_REMOTE_FOLDER'] + "_agent.log", log_file]
        logger.debug("Copy agent log from %s@%s: %s", self.username, self.host, cmd)
        remove = self.ssh.get_scp_pipe(cmd)
        remove.wait()

        logger.info("Removing agent from: %s@%s...", self.username, self.host)
        cmd = ['rm', '-r', self.path['AGENT_REMOTE_FOLDER']]
        remove = self.ssh.get_ssh_pipe(cmd)
        remove.wait()
        return log_file


class MonitoringCollector:

    """Aggregate data from several collectors"""

    def __init__(self):
        self.config = None
        self.default_target = None
        self.agents = []
        self.agent_pipes = []
        self.filter_conf = {}
        self.listeners = []
        self.ssh_wrapper_class = SSHWrapper
        self.first_data_received = False
        self.send_data = ''
        self.artifact_files = []
        self.inputs, self.outputs, self.excepts = [], [], []
        self.filter_mask = defaultdict(str)
        self.ssh_timeout = 5
        self.load_start_time = None

    def add_listener(self, obj):
        self.listeners.append(obj)

    def prepare(self):

        """Prepare for monitoring - install agents etc"""

        # Parse config
        agent_config = []
        if self.config:
            [agent_config, self.filter_conf] = self.getconfig(self.config, self.default_target)

        logger.debug("filter_conf: %s", self.filter_conf)

        # Filtering
        for host in self.filter_conf:
            self.filter_mask[host] = []
        logger.debug("Filter mask: %s", self.filter_mask)

        # Creating agent for hosts
        logger.debug('Creating agents')
        for adr in agent_config:
            logger.debug('Creating agent: %s', adr)
            agent = AgentClient()
            agent.host = adr['host']
            agent.username = adr['username']
            agent.python = adr['python']
            agent.metric = adr['metric']
            agent.port = adr['port']
            agent.interval = adr['interval']
            agent.custom = adr['custom']
            agent.startups = adr['startups']
            agent.shutdowns = adr['shutdowns']
            agent.ssh = self.ssh_wrapper_class(self.ssh_timeout)
            self.agents.append(agent)

        # Mass agents install
        logger.debug("Agents: %s", self.agents)

        conf = Config(self.config)
        for agent in self.agents:
            logger.debug('Install monitoring agent. Host: %s', agent.host)
            self.artifact_files.append(agent.install(conf.loglevel()))

    def start(self):
        """Start N parallel agents"""
        for agent in self.agents:
            pipe = agent.start()
            self.agent_pipes.append(pipe)

            fds = pipe.stdout.fileno()
            flags = fcntl.fcntl(fds, fcntl.F_GETFL)
            fcntl.fcntl(fds, fcntl.F_SETFL, flags | os.O_NONBLOCK)
            self.outputs.append(pipe.stdout)

            fds = pipe.stderr.fileno()
            flags = fcntl.fcntl(fds, fcntl.F_GETFL)
            fcntl.fcntl(fds, fcntl.F_SETFL, flags | os.O_NONBLOCK)
            self.excepts.append(pipe.stderr)

        logger.debug("Pipes: %s", self.agent_pipes)

    def poll(self):
        """Poll agents for data"""
        readable, writable, exceptional = select.select(self.outputs, self.inputs, self.excepts, 0)
        logger.debug("Streams: %s %s %s", readable, writable, exceptional)

        # if empty run - check children
        if (not readable) or exceptional:
            for pipe in self.agent_pipes:
                if pipe.returncode:
                    logger.debug("Child died returncode: %s", pipe.returncode)
                    self.outputs.remove(pipe.stdout)
                    self.agent_pipes.remove(pipe)

        # Handle exceptions
        for excepted in exceptional:
            data = excepted.readline()
            while data:
                logger.error("Got exception [%s]: %s", excepted, data)
                data = excepted.readline()

        while readable:
            to_read = readable.pop(0)
            # Handle outputs

            try:
                lines = to_read.read().split("\n")
            except IOError:
                logger.debug("No data available")
                lines = []

            for data in lines:
                logger.debug("Got data from agent: %s", data.strip())
                self.send_data += self.filter_unused_data(self.filter_conf, self.filter_mask, data)
                logger.debug("Data after filtering: %s", self.send_data)

        if not self.first_data_received and self.send_data:
            self.first_data_received = True
            logger.info("Monitoring received first data")
        else:
            self.send_collected_data()

        return len(self.outputs)

    def stop(self):
        """Shutdown agents"""
        logger.debug("Initiating normal finish")
        for pipe in self.agent_pipes:
            try:
                pipe.stdin.write("stop\n")
            except IOError as exc:
                logger.warn("Problems stopping agent: %s", traceback.format_exc(exc))

        time.sleep(1)

        for pipe in self.agent_pipes:
            if pipe.pid:
                first_try = True
                delay = 1
                while tankcore.pid_exists(pipe.pid):
                    if first_try:
                        logger.debug("Killing %s with %s", pipe.pid, signal.SIGTERM)
                        os.killpg(pipe.pid, signal.SIGTERM)
                        pipe.communicate()
                        first_try = False
                        time.sleep(0.1)
                    else:
                        time.sleep(delay)
                        delay *= 2
                        logger.warn("Killing %s with %s", pipe.pid, signal.SIGKILL)
                        os.killpg(pipe.pid, signal.SIGKILL)

        for agent in self.agents:
            self.artifact_files.append(agent.uninstall())

    def send_collected_data(self):
        """sends pending data set to listeners"""
        for listener in self.listeners:
            listener.monitoring_data(self.send_data)
        self.send_data = ''

    def get_host_config(self, host, names, target_hint):

        default = {
            'System': 'csw,int',
            'CPU': 'user,system,iowait',
            'Memory': 'free,cached,used',
            'Disk': 'read,write',
            'Net': 'recv,send',
        }

        default_metric = ['CPU', 'Memory', 'Disk', 'Net']

        hostname = host.get('address').lower()
        if hostname == '[target]':
            if not target_hint:
                raise ValueError("Can't use [target] keyword with no target parameter specified")
            logger.debug("Using target hint: %s", target_hint)
            hostname = target_hint.lower()
        stats = []
        startups = []
        shutdowns = []
        custom = {'tail': [], 'call': [], }
        metrics_count = 0
        for metric in host:
            # known metrics
            if metric.tag in default.keys():
                metrics_count += 1
                metr_val = default[metric.tag].split(',')
                if metric.get('measure'):
                    metr_val = metric.get('measure').split(',')
                for elm in metr_val:
                    if not elm:
                        continue
                    stat = "%s_%s" % (metric.tag, elm)
                    stats.append(stat)
                    agent_name = self.get_agent_name(metric.tag, elm)
                    if agent_name:
                        names[agent_name] = 1
            # custom metric ('call' and 'tail' methods)
            elif (str(metric.tag)).lower() == 'custom':
                metrics_count += 1
                isdiff = metric.get('diff')
                if not isdiff:
                    isdiff = 0
                stat = "%s:%s:%s" % (base64.b64encode(metric.get('label')), base64.b64encode(metric.text), isdiff)
                stats.append('Custom:' + stat)
                custom[metric.get('measure', 'call')].append(stat)
            elif (str(metric.tag)).lower() == 'startup':
                startups.append(metric.text)
            elif (str(metric.tag)).lower() == 'shutdown':
                shutdowns.append(metric.text)

        logger.debug("Metrics count: %s", metrics_count)
        logger.debug("Host len: %s", len(host))
        logger.debug("keys: %s", host.attrib.keys())
        logger.debug("values: %s", host.attrib.values())

        # use default metrics for host
        if metrics_count == 0:
            for metric in default_metric:
                metr_val = default[metric].split(',')
                for elm in metr_val:
                    stat = "%s_%s" % (metric, elm)
                    stats.append(stat)
                    agent_name = self.get_agent_name(metric, elm)
                    if agent_name:
                        names[agent_name] = 1

        metric = ','.join(names.keys())

        return {
            'metric': metric or 'cpu-stat',
            'interval': host.get('interval', 1),
            'priority': host.get('priority', 0),
            'port': host.get('port', '22'),
            'python': host.get('python', '/usr/bin/env python2'),
            'username': host.get('username', getpass.getuser()),
            'custom': custom,
            'host': hostname,
            'startups': startups,
            'shutdowns': shutdowns,

            # XXX: should be separate?
            'stats': {hostname: stats},
        }

    def getconfig(self, filename, target_hint):
        """Prepare config data"""

        try:
            tree = parse_xml(filename)
        except IOError as exc:
            logger.error("Error loading config: %s", exc)
            raise RuntimeError("Can't read monitoring config %s" % filename)

        hosts = tree.findall('Host')
        names = defaultdict()
        config = []

        filter_obj = defaultdict(str)
        for host in hosts:
            host_config = self.get_host_config(host, names, target_hint)
            # XXX: why stats should be separated?
            filter_obj.update(host_config.pop('stats'))
            config.append(host_config)

        return [config, filter_obj]

    def filtering(self, mask, filter_list):
        """Filtering helper"""
        host = filter_list[0]
        initial = [0, 1]
        res = []
        if mask[host]:
            keys = initial + mask[host]
            for key in keys:
                try:
                    res.append(filter_list[key])
                except IndexError:
                    logger.warn("Problems filtering data: %s with %s", mask, len(filter_list))
                    return None
        return ';'.join(res)

    def filter_unused_data(self, filter_conf, filter_mask, data):
        """Filter unselected metrics from data"""
        logger.debug("Filtering data: %s", data)
        out = ''
        # Filtering data
        keys = data.rstrip().split(';')
        if re.match('^start;', data):  # make filter_conf mask
            host = keys[1]
            for i in range(3, len(keys)):
                if keys[i] in filter_conf[host]:
                    filter_mask[host].append(i - 1)
            logger.debug("Filter mask: %s", filter_mask)
            out = 'start;'
            out += self.filtering(filter_mask, keys[1:]).rstrip(';') + '\n'
        elif re.match('^\[debug\]', data):  # log debug output
            logger.debug('agent debug: %s', data.rstrip())
        else:
            # if we are in start_test() phase, check data's timestamp with load_start_time
            # and skip data collected before load_start_time
            if self.load_start_time is not None:
                try:
                    if int(keys[1]) >= self.load_start_time:
                        filtered = self.filtering(filter_mask, keys)
                        if filtered:
                            out = filtered + '\n'  # filtering values
                except IndexError:
                    pass
        return out

    def get_agent_name(self, metric, param):
        """Resolve metric name"""
        depend = {
            'CPU': {
                'idle': 'cpu-stat',
                'user': 'cpu-stat',
                'system': 'cpu-stat',
                'iowait': 'cpu-stat',
                'nice': 'cpu-stat'
            },
            'System': {
                'la1': 'cpu-la',
                'la5': 'cpu-la',
                'la15': 'cpu-la',
                'csw': 'cpu-stat',
                'int': 'cpu-stat',
                'numproc': 'cpu-stat',
                'numthreads': 'cpu-stat',
            },
            'Memory': {
                'free': 'mem',
                'used': 'mem',
                'cached': 'mem',
                'buff': 'mem',
            },
            'Disk': {
                'read': 'disk',
                'write': 'disk',
            },
            'Net': {
                'recv': 'net',
                'send': 'net',
                'tx': 'net-tx-rx',
                'rx': 'net-tx-rx',
                'retransmit': 'net-retrans',
                'estab': 'net-tcp',
                'closewait': 'net-tcp',
                'timewait': 'net-tcp',
            }
        }
        if depend[metric][param]:
            return depend[metric][param]
        else:
            return ''


class MonitoringDataListener:
    """Parent class for data listeners"""

    def __init__(self):
        pass

    def monitoring_data(self, data_string):
        """Notification about new monitoring data lines"""
        raise NotImplementedError()


class StdOutPrintMon(MonitoringDataListener):
    """Simple listener, writing data to stdout"""

    def __init__(self):
        MonitoringDataListener.__init__(self)

    def monitoring_data(self, data_string):
        sys.stdout.write(data_string)


class MonitoringDataDecoder:
    """The class that serves converting monitoring data lines to dict"""
    NA = 'n/a'

    def __init__(self):
        self.metrics = {}

    def decode_line(self, line):
        """convert mon line to dict"""
        is_initial = False
        data_dict = {}
        data = line.strip().split(';')
        timestamp = -1
        if data[0] == 'start':
            data.pop(0)  # remove 'start'
            host = data.pop(0)
            if not data:
                logger.warn("Wrong mon data line: %s", line)
            else:
                timestamp = data.pop(0)
                self.metrics[host] = []
                for metric in data:
                    if metric.startswith("Custom:"):
                        metric = base64.standard_b64decode(metric.split(':')[1])
                    self.metrics[host].append(metric)
                    data_dict[metric] = self.NA
                    is_initial = True
        else:
            host = data.pop(0)
            timestamp = data.pop(0)

            if host not in self.metrics.keys():
                raise ValueError("Host %s not in started metrics: %s" % (host, self.metrics))

            if len(self.metrics[host]) != len(data):
                raise ValueError("Metrics len and data len differs: %s vs %s" % (len(self.metrics[host]), len(data)))

            for metric in self.metrics[host]:
                data_dict[metric] = data.pop(0)

        logger.debug("Decoded data %s: %s", host, data_dict)
        return host, data_dict, is_initial, timestamp


# FIXME: 3 synchronize times between agent and collector better



''' Plugin provides fullscreen console '''
import logging
import sys
import traceback
import threading

from Aggregator import AggregatorPlugin, AggregateResultListener
from ConsoleScreen import Screen
from yandextank.core import AbstractPlugin


class ConsoleOnlinePlugin(AbstractPlugin, AggregateResultListener):
    ''' Console plugin '''
    SECTION = 'console'

    def __init__(self, core):
        AbstractPlugin.__init__(self, core)
        self.screen = None
        self.render_exception = None
        self.console_markup = None
        self.remote_translator = None
        self.info_panel_width = '33'
        self.short_only = 0
        # these three provide non-blocking console output
        self.__console_view = None
        self.__writer_thread = None
        self.__writer_event = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return ["info_panel_width", "short_only", "disable_all_colors", "disable_colors"]

    def configure(self):
        self.info_panel_width = self.get_option("info_panel_width", self.info_panel_width)
        self.short_only = int(self.get_option("short_only", '0'))
        if not int(self.get_option("disable_all_colors", '0')):
            self.console_markup = RealConsoleMarkup()
        else:
            self.console_markup = NoConsoleMarkup()
        for color in self.get_option("disable_colors", '').split(' '):
            self.console_markup.__dict__[color] = ''
        self.screen = Screen(self.info_panel_width, self.console_markup)

        try:
            aggregator = self.core.get_plugin_of_type(AggregatorPlugin)
            aggregator.add_result_listener(self)
        except KeyError:
            self.log.debug("No aggregator for console")
            self.screen.block_rows = []
            self.screen.info_panel_percent = 100

    def __console_writer(self):
        while True:
          self.__writer_event.wait()
          self.__writer_event.clear()

          if self.__console_view:
              if not self.short_only:
                  self.log.debug("Writing console view to STDOUT")
                  sys.stdout.write(self.console_markup.clear)
                  sys.stdout.write(self.__console_view)
                  sys.stdout.write(self.console_markup.TOTAL_RESET)

              if self.remote_translator:
                  self.remote_translator.send_console(self.__console_view)

    def is_test_finished(self):
        if not self.__writer_thread:
          self.__writer_event = threading.Event()
          self.__writer_thread = threading.Thread(target=self.__console_writer)
          self.__writer_thread.daemon = True
          self.__writer_thread.start()

        try:
            self.__console_view = self.screen.render_screen().encode('utf-8')
        except Exception, ex:
            self.log.warn("Exception inside render: %s", traceback.format_exc(ex))
            self.render_exception = ex
            self.__console_view = ""

        self.__writer_event.set()
        return -1


    def aggregate_second(self, second_aggregate_data):
        self.screen.add_second_data(second_aggregate_data)
        if self.short_only:
            tpl = "Time: %s\tExpected RPS: %s\tActual RPS: %s\tActive Threads: %s\tAvg RT: %s"
            ovr = second_aggregate_data.overall  # just to see the next line in IDE
            data = (second_aggregate_data.time, ovr.planned_requests, ovr.RPS,
                    ovr.active_threads, ovr.avg_response_time)
            self.log.info(tpl % data)


    def add_info_widget(self, widget):
        ''' add right panel widget '''
        if not self.screen:
            self.log.debug("No screen instance to add widget")
        else:
            self.screen.add_info_widget(widget)


# ======================================================


class RealConsoleMarkup(object):
    '''
    Took colors from here: https://www.siafoo.net/snippet/88
    '''
    WHITE_ON_BLACK = '\033[37;40m'
    TOTAL_RESET = '\033[0m'
    clear = "\x1b[2J\x1b[H"
    new_line = u"\n"

    YELLOW = '\033[1;33m'
    RED = '\033[1;31m'
    RED_DARK = '\033[31;3m'
    RESET = '\033[1;m'
    CYAN = "\033[1;36m"
    GREEN = "\033[1;32m"
    WHITE = "\033[1;37m"
    MAGENTA = '\033[1;35m'
    BG_MAGENTA = '\033[1;45m'
    BG_GREEN = '\033[1;42m'
    BG_BROWN = '\033[1;43m'
    BG_CYAN = '\033[1;46m'

    def clean_markup(self, orig_str):
        ''' clean markup from string '''
        for val in [self.YELLOW, self.RED, self.RESET,
                    self.CYAN, self.BG_MAGENTA, self.WHITE,
                    self.BG_GREEN, self.GREEN, self.BG_BROWN,
                    self.RED_DARK, self.MAGENTA, self.BG_CYAN]:
            orig_str = orig_str.replace(val, '')
        return orig_str


# ======================================================
# FIXME: 3 better way to have it?

class NoConsoleMarkup(RealConsoleMarkup):
    ''' all colors are disabled '''
    WHITE_ON_BLACK = ''
    TOTAL_RESET = ''
    clear = ""
    new_line = u"\n"

    YELLOW = ''
    RED = ''
    RED_DARK = ''
    RESET = ''
    CYAN = ""
    GREEN = ""
    WHITE = ""
    MAGENTA = ''
    BG_MAGENTA = ''
    BG_GREEN = ''
    BG_BROWN = ''
    BG_CYAN = ''


# ======================================================


class AbstractInfoWidget:
    ''' parent class for all right panel widgets '''

    def __init__(self):
        self.log = logging.getLogger(__name__)

    def render(self, screen):
        raise NotImplementedError()

    def get_index(self):
        ''' get vertical priority index '''
        return 0



import time
from collections import namedtuple
from yandextank.core import AbstractPlugin
import logging
from random import randint
import threading as th
import sys

from sqlalchemy import create_engine
from sqlalchemy import exc

import requests

requests_logger = logging.getLogger('requests')
requests_logger.setLevel(logging.WARNING)

Sample = namedtuple(
    'Sample', 'marker,threads,overallRT,httpCode,netCode,sent,received,connect,send,latency,receive,accuracy')

from contextlib import contextmanager


@contextmanager
def measure(marker, results):
    start_time = time.time()
    yield
    response_time = int((time.time() - start_time) * 1000)
    data_item = Sample(
        marker,             # marker
        th.active_count(),  # threads
        response_time,      # overall response time
        200,                # httpCode
        0,                  # netCode
        0,                  # sent
        0,                  # received
        0,                  # connect
        0,                  # send
        response_time,      # latency
        0,                  # receive
        0,                  # accuracy
    )
    results.put((int(time.time()), data_item), timeout=1)


class LogGun(AbstractPlugin):
    SECTION = 'log_gun'

    def __init__(self, core):
        self.log = logging.getLogger(__name__)
        AbstractPlugin.__init__(self, core)
        param = self.get_option("param", '15')
        self.log.info('Initialized log gun for BFG with param = %s' % param)

    def shoot(self, missile, marker, results):
        self.log.info("Missile: %s\n%s", marker, missile)
        rt = randint(2, 30000)
        data_item = Sample(
            marker,             # marker
            th.active_count(),  # threads
            rt,                 # overallRT
            0,                  # httpCode
            0,                  # netCode
            0,                  # sent
            0,                  # received
            0,                  # connect
            0,                  # send
            rt,                 # latency
            0,                  # receive
            0,                  # accuracy
        )
        results.put((int(time.time()), data_item), timeout=1)

class SqlGun(AbstractPlugin):
    SECTION = 'sql_gun'

    def __init__(self, core):
        self.log = logging.getLogger(__name__)
        AbstractPlugin.__init__(self, core)
        self.engine = create_engine(self.get_option("db"))

    def shoot(self, missile, marker, results):
        self.log.debug("Missile: %s\n%s", marker, missile)
        start_time = time.time()
        errno = 0
        httpCode = 200
        try:
            cursor = self.engine.execute(missile.replace('%', '%%'))
            cursor.fetchall()
            cursor.close()
        except exc.TimeoutError as e:
            self.log.debug("Timeout: %s", e)
            errno = 110
        except exc.ResourceClosedError as e:
            self.log.debug(e)
        except exc.SQLAlchemyError as e:
            httpCode = 500
            self.log.debug(e.orig.args)
        except exc.SAWarning as e:
            httpCode = 400
            self.log.debug(e)
        except Exception as e:
            httpCode = 500
            self.log.debug(e)
        rt = int((time.time() - start_time) * 1000)
        data_item = Sample(
            marker,             # marker
            th.active_count(),  # threads
            rt,                 # overallRT
            httpCode,           # httpCode
            errno,              # netCode
            0,                  # sent
            0,                  # received
            0,                  # connect
            0,                  # send
            rt,                 # latency
            0,                  # receive
            0,                  # accuracy
        )
        results.put((int(time.time()), data_item), timeout=1)

class CustomGun(AbstractPlugin):
    SECTION = 'custom_gun'

    def __init__(self, core):
        self.log = logging.getLogger(__name__)
        AbstractPlugin.__init__(self, core)
        module_path = self.get_option("module_path", "")
        module_name = self.get_option("module_name")
        if module_path:
            sys.path.append(module_path)
        self.module = __import__(module_name)

    def shoot(self, missile, marker, results):
        self.module.shoot(missile, marker, results)

class HttpGun(AbstractPlugin):
    SECTION = 'http_gun'

    def __init__(self, core):
        self.log = logging.getLogger(__name__)
        AbstractPlugin.__init__(self, core)
        self.base_address = self.get_option("base_address")

    def shoot(self, missile, marker, results):
        self.log.debug("Missile: %s\n%s", marker, missile)
        self.log.debug("Sending request: %s", self.base_address + missile)
        start_time = time.time()
        r = requests.get(self.base_address + missile)
        errno = 0
        httpCode = r.status_code
        rt = int((time.time() - start_time) * 1000)
        data_item = Sample(
            marker,             # marker
            th.active_count(),  # threads
            rt,                 # overallRT
            httpCode,           # httpCode
            errno,              # netCode
            0,                  # sent
            0,                  # received
            0,                  # connect
            0,                  # send
            rt,                 # latency
            0,                  # receive
            0,                  # accuracy
        )
        results.put((int(time.time()), data_item), timeout=1)


class ScenarioGun(AbstractPlugin):
    SECTION = 'scenario_gun'

    def __init__(self, core):
        self.log = logging.getLogger(__name__)
        AbstractPlugin.__init__(self, core)
        module_path = self.get_option("module_path", "")
        module_name = self.get_option("module_name")
        if module_path:
            sys.path.append(module_path)
        self.module = __import__(module_name)
        self.scenarios = self.module.SCENARIOS

    def shoot(self, missile, marker, results):
        scenario = self.scenarios.get(marker, None)
        if scenario:
            try:
                scenario(missile, marker, results)
            except RuntimeError as e:
                self.log.warning("Scenario %s failed with %s", marker, e)
        else:
            self.log.warning("Scenario not found: %s", marker)
